
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */

/* *************************************************************

                        P I X C O N

This program is a 3D renderer.  It is acivated by the following
command at the command line.


        pixcon <-v filename> <-l filename> <-a filename>

        -v filename : user defined file to represent camera
                      default is "object.vew"

        -l filename : user defined file to list object data
                      default is "object.lst"

        -a filename : user define file to anti-aliasing filter
                      default is internal, and is only activated
                      if the -a option is invoked w/o a filename

The program requires 5 basic types of data files - camera specification,
object list, object structure, object color, and antialiasing filter.

   The first data file is the camera data in the following form:

        X Y Z                   - camera location
        X Y Z                   - camera focus point
        X Y Z                   - orientation of the camera's Y axis
        D                       - distance from camera to viewplane
        W                       - width of viewplane
        X Y Z                   - location of light source
        R B G                   - ambiant R B G lighting
        R B G                   - point light source R B G reflection

   The data file that lists the objects to be rendered is of the
   following form:

        # of objects, n

        object 1 type (POLYGON, SPATCH)
        object 1 coloring (BW, CONSTANT, GOURAUD, PHONG, PHONG2)
        object 1 object data file
        object 1 color file
        object 1 transformation matrix (4X4)
        object 1 scale factor

                - - - - -

        object n type (POLYGON, SPATCH)
        object n coloring type (BW, CONSTANT, GOURAUD, PHONG, PHONG2)
        object n object data file
        object n color file
        object n transformation matrix (4X4)
        object n scale factor

   The file structure for polygonal object data is:

   M                                    - # of vertices   m
   N                                    - # of polygons   n
   X1 Y1 Z1                             - x y z coordinates for vertex 1
    - - -
   XM YM ZM                             - x y z coordinates for vertex N
   C1 V1 ... Vc1                        - # of vertices in this polygon, and
    - - -                                   pointers to vertex list...
   CN VN ... VcN

   The file structure for the coloring data for polygons is:

                                        - general color data
        descriptor KAR KAB KAG          - ambiant color ( range 0..1 )
        descriptor KPR KPB KPG          - point color ( range 0..1 )
        descriptor KSR KSB KSG          - specularization color ( range 0..1 )
        descriptor FAR FAB FAG          - attenuation ( range 0..1 )
        descriptor SPEC_N               - power of cos

        N                               - # of polygons whos color differs
                                          from rest of object

        X1                              - polygon number
        descriptor KAR1 KAB1 KAG1       - ambiant color ( range 0..1 )
        descriptor KPR1 KPB1 KPG1       - point color ( range 0..1 )
        descriptor KSR1 KSB1 KSG1       - specularization color ( range 0..1 )
        descriptor FAR1 FAB1 FAG1       - attenuation ( range 0..1 )
        descriptor SPEC_N1              - power of cos
                - - -
        XN                              - polygon number
        descriptor KARN KABN KAGN       - ambiant color ( range 0..1 )
        descriptor KPRN KPBN KPGN       - point color ( range 0..1 )
        descriptor KSRN KSBN KSGN       - specularization color ( range 0..1 )
        descriptor FARN FABN FAGN       - attenuation ( range 0..1 )
        descriptor SPEC_NN              - power of cos

   The file structure for the anti-aliasing filter is a simple
   2D filter.  If smaller filters are to be used, pad the edges of
   the filter with zeros to reach the 7x7 requirement.

written by:     Alex Harper
                04/01/93
                CS 219
************************************************************* */

#include "pixcon.h"
#include <string.h>
#include "laser.h"

/* **************************************************
   main: program controller
************************************************** */
void main (int argc, char *argv[]) {

   listtype     mainlist;               // ptr to objects

   camera       cparm;                  // camera data
   light        lparm;
   int          mapcount;               // # of pixmaps
   int          i, j, k;                   // looping variables
   slave        *ptr;
   char         *queue = (char *)NULL;

   mapul        *mapbuffer;
   int          flagscreengl;
   texbase	*tptr;
   FILE         *infile, *outfile;

   time_t	oldtime, newtime;

   GLint attribs[] = {
                GLX_RGBA,
                GLX_DEPTH_SIZE, 24,
                GLX_DOUBLEBUFFER,
                None
   };



#ifdef SGI
/*
//   sigset(SIGCLD, (void (*)(int ... ))killchild);
   sigset(SIGINT, (void (*)(int ... ))killchild);
   sigset(SIGKILL, (void (*)(int ... ))killchild);
   sigset(SIGTERM, (void (*)(int ... ))killchild);
*/
   sigset(SIGINT, (void (*)( ... ))killchild);
   sigset(SIGKILL, (void (*)( ... ))killchild);
   sigset(SIGTERM, (void (*)( ... ))killchild);
#else
   signal(SIGINT, kc);
   signal(SIGKILL, kc);
   signal(SIGTERM, kc);
#endif

   printf("Parsing command-line arguments...\n");
   parseinput(argc, argv);			// parse command line

   if (!SILENT)
      printf("Parsing shaders...\n");

   setup_shader();
   init_trig();

   if (DEFCAMERA)                            // read camera data
      infile = fopen(SDEFCAMERA, "r");
   else
      infile = fopen("object.vew", "r");

   if (infile) {
      cparm.read_def(infile);
      fclose(infile);
   }

   else
      cparm.load_def();

   if (DEFLIGHT)                            // read light data
      infile = fopen(SDEFLIGHT, "r");
   else
      infile = fopen("object.lit", "r");

   if (infile) {
      lparm.read_data(infile, &mainlist.tdtr);
      fclose(infile);
   }

   if (!SILENT)
      printf("Parsing .lst file...\n");

   if (DEFLST) {                           // read in object data
      if (!parse_list(SDEFLST, &mainlist, &mapcount, &cparm, shades))
         read_list(SDEFLST, &mainlist, &mapcount, &cparm, shades);
   }

   else
      if (!parse_list("object.lst", &mainlist, &mapcount, &cparm, shades))
         read_list("object.lst", &mainlist, &mapcount, &cparm, shades);

   if (!SILENT)
      printf("Processing mipmaps...\n");

   if (!NIXMIP)
      for (tptr=mainlist.tdtr; tptr != (texbase *)NULL; tptr=tptr->next)
         if (tptr->query_whatami() == TEXTURE)
            ((texture *)tptr)->build_mipmaps();

   if (SAVE) {  			// save file
      outfile = fopen(SSAVE, "wb");
      fwrite(&mapcount, sizeof(int), 1, outfile);
      fwrite(&WINY, sizeof(int), 1, outfile);
      fwrite(&WINX, sizeof(int), 1, outfile);
   }

   if (!NOSCREEN)
      gwindow.AOGLreset(argc, argv, WINX, WINY, "Pixcon", attribs);

   flagscreengl = (!(NOSCREEN || PENCILTEST));

   if (flagscreengl)
      mapbuffer = new mapul[mapcount];

   queue = new char[mapcount];

   if (RENDER) {
      if (LQ4000_init("/dev/ttyd2")) {
         printf("Could not open laser disk...\n");
         exit(-1);
      }

      if (LQ4000_rec_standby(RENDER_START, RENDER_END)) {
         printf("Bad start/stop frames...\n");
         exit(-1);
      }

   }

   control.activate();

   if (!SILENT)
      printf("Start Rendering...\n");

   time(&oldtime);

   j = i = k = 0;
   while (j < mapcount) {

      if (flagscreengl && k < mapcount) {
         mapbuffer[k].init_map(WINX, WINY);
         k++;
      }

      if (i < mapcount && control.query_free()) {
         ptr = control.get_slave();

         if (mainlist.camdtr[i] == (camera *)NULL) {
            mainlist.camdtr[i] = new camera;
            *mainlist.camdtr[i] = cparm;
         }

         if (mainlist.lightdtr[i] == (light *)NULL) {
            mainlist.lightdtr[i] = new light;
            *mainlist.lightdtr[i] = lparm;
         }

         if (flagscreengl)
            ptr->spawn(mainlist.camdtr[i], mainlist.lightdtr[i], outfile, &mainlist, i, &mapbuffer[i]);
         else
            ptr->spawn(mainlist.camdtr[i], mainlist.lightdtr[i], outfile, &mainlist, i, (mapul *)NULL);

         queue[i] = 0;
         i++;
      }

      if (control.result != (ptrtype *)NULL) {
         control.set_lock(LOCK_RESULT);
         queue[control.result->id] = 1;
         control.pop_result();
         control.free_lock(LOCK_RESULT);
      }

      if (queue[j]) {

         if (flagscreengl) {
            if (mapcount > 1)
               gwindow.AOGLbitmap(&mapbuffer[j]);
         }

         else
            if (PENCILTEST)
               gwindow.AOGLswapbuffers();

//         if (j)
//            gwindow.ALEXpause();

//         printf("%d freeze\n", j);
//         gwindow.ALEXfreezeframe(&mapbuffer[j]);

         if (RENDER)
            if (LQ4000_rec()) {
               printf("Laser disk error...\n");
               exit(-1);
            }

         if (!SILENT)
            printf("FINISHED %d :)\n", j);

         j++;
      }

   }

   time(&newtime);

   control.done();
   control.kill_children();

   if (SAVE)					// save file
      fclose(outfile);

   if (!SILENT) {
      printf("######  ####### #     # #######\n");
      printf("#     # #     # ##    # #\n");
      printf("#     # #     # # #   # #\n");
      printf("#     # #     # #  #  # #####\n");
      printf("#     # #     # #   # # #\n");
      printf("#     # #     # #    ## #\n");
      printf("######  ####### #     # #######\n");

      printf("Average rendering time = %f\n", (newtime - oldtime)/((float)mapcount));
      printf("Have a nice day :)\n");
   }

   if (PENCILTEST) 
      gwindow.ALEXpause();
   else
      if (flagscreengl)
         if (mapcount == 1)
            gwindow.AOGLbitblt(mapbuffer);                       // pixcon
         else
            gwindow.AOGLanimate(mapcount, mapbuffer);            // animaron && reannie

   if (RENDER)
      LQ4000_close();

   if (!NOSCREEN)
      gwindow.ALEXcancel();		// independent of OGL/X
}
